/*
Module : Mailslot.h
Purpose: Defines the interface for a C++ wrapper class for Win32 Mailslots
Created: PJN / 18-06-1997
History: PJN / 25-07-1998 1. Provision of Unicode build configurations
                          2. Updated make file to VC 5 format
                          3. Fixed a bug in CClientMailslot destructor
                          4. CServerMailslot::Close() & CClientMailslot::Close() now work correctly 
                          when the mailslot is already closed
                          5. Fixed a level 4 warning in the CWinNotifyDoc in the	demo program.
                          6. Changed 2 functions which used a pointer parameter to now use a C++ reference instead.
         PJN / 10-08-1998 1. Close function of both classes now correctly sets 
                          mailslot handle back to INVALID_HANDLE_VALUE
         PJN / 26-07-2003 1. Updated copyright details.
                          2. Fixed issue in CClientMailslot::Open which was causing it to only allow one 
                          instance of the mailslot to be opened. Thanks to GdP Software for reporting this problem.
         PJN / 22-12-2006 1. Updated copyright details
                          2. Updated the documentation to use the same style as the web site
                          3. Optimized CServerMailslot constructor code
                          4. Addition of a CMAILSLOT_EXT_CLASS preprocessor macro to allow the classes to be more 
                          easily added to an extension dll
                          5. Reviewed all TRACE statements for correctness
                          6. Optimized CClientMailslot constructor code
                          7. Updated the code to clean compile on VC 2005
                          8. Code now uses newer C++ style casts instead of C style casts.
                          9. Changed the default size of the formview window in the sample ap to avoid scroll bars
                          using default fonts.
         PJN / 23-12-2006 1. Code now uses CString instead of raw TCHAR buffers in various functions
                          2. Classes now do not derive from CObject as it was not really required
         PJN / 08-09-2007 1. Updated copyright details.
                          2. Minor code tidy up.
         PJN / 07-09-2008 1. Updated copyright details.
                          2. Code now compiles cleanly using Code Analysis (/analyze)
                          3. Updated code to compile correctly using _ATL_CSTRING_EXPLICIT_CONSTRUCTORS define
                          4. Updated sample app to clean compile on VC 2008
                          5. The code has now been updated to support VC 2005 or later only. 
                          6. Removed VC 6 style AppWizard comments from the code.
                          7. Updated the sample app to improve on the error reporting.
         PJN / 10-05-2015 1. Updated copyright details.
                          2. Updated the project settings to more modern defaults
                          3. Removed all unnecessary TRACE calls
                          4. Made the classes into a header only implementation
                          5. Updated the code to clean compile on VC 2013
                          6. Reworked the CServerMailslot::Open and CClientMailslot::Open to provide the full mailslot 
                          name as its first parameter. This allows the class to avoid the use of CString
                          7. Addition of a CServerMailslot::GetInfo method which encapsulates GetMailslotInfo.
                          8. Removed the CServerMailslot::MessageWaiting, MessageCount, SizeOfWaitingMessage,  
                          MaximumMessageSize, GetCreationTime, GetLastAccessTime & GetLastWriteTime methods
                          9. Added SAL annotations to all the code
                          10. Reworked the CServerMailslot::Read method to use a pointer for returning the number of
                          bytes read
                          11. Reworked the CClientMailslot::Write method to use a pointer for returning the number of
                          bytes write
                          12. Replaced all ASSERT calls with ATLASSERT. This means that the class can now be used
                          without the need to pull in MFC
         PJN / 16-01-2016 1. Updated copyright details.
                          2. Verified the code compiles cleanly in VC 2015
         PJN / 30-04-2017 1. Updated copyright details.
                          2. Updated the code to compile cleanly using /permissive-.
         PJN / 14-10-2018 1. Updated copyright details.
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes mean that the code will 
                          now only compile on VC 2017 or later.
         PJN / 03-06-2019 1. Updated copyright details.
                          2. Updated the code to clean compile on VC 2019
         PJN / 26-12-2019 1. Fixed various Clang-Tidy static code analysis warnings in the code.
                          2. Replaced BOOL with bool throughout the code.
         PJN / 03-05-2020 1. Updated copyright details.
                          2. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 09-01-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable declarations

Copyright (c) 1997 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


/////////////////////////////// Macros / Defines //////////////////////////////

#pragma once

#ifndef __MAILSLOT_H__
#define __MAILSLOT_H__

#ifndef CMAILSLOT_EXT_CLASS
#define CMAILSLOT_EXT_CLASS
#endif //#ifndef CMAILSLOT_EXT_CLASS


/////////////////////////////// Classes ///////////////////////////////////////

class CMAILSLOT_EXT_CLASS CServerMailslot
{
public:
//Constructors / Destructors
  CServerMailslot() noexcept : m_hMailslot{INVALID_HANDLE_VALUE}
  {
  }
  CServerMailslot(const CServerMailslot&) = delete;
  CServerMailslot(CServerMailslot&&) = delete;
  ~CServerMailslot()
  {
    Close();
  }

//Operations
  CServerMailslot& operator=(const CServerMailslot&) = delete;
  CServerMailslot& operator=(CServerMailslot&&) = delete;

  bool Open(_In_ LPCTSTR lpszName, _In_ DWORD dwMaxMessageSize = 0, _In_ DWORD dwReadTimeout = 0, _In_opt_ LPSECURITY_ATTRIBUTES lpSecurityAttributes = nullptr) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(!IsOpen());

    m_hMailslot = CreateMailslot(lpszName, dwMaxMessageSize, dwReadTimeout, lpSecurityAttributes);
    return (m_hMailslot != INVALID_HANDLE_VALUE);
  }

  bool Close() noexcept
  {
    //What will be the return value from this method (assume the best)
    bool bSuccess{true};

    if (IsOpen())
    {
      bSuccess = CloseHandle(m_hMailslot);
      m_hMailslot = INVALID_HANDLE_VALUE;
    }
    return bSuccess;
  }

  bool Read(_Out_writes_bytes_to_opt_(nNumberOfBytesToRead, *lpNumberOfBytesRead) __out_data_source(FILE) LPVOID lpBuffer, _In_ DWORD nNumberOfBytesToRead, _Out_opt_ LPDWORD lpNumberOfBytesRead) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(IsOpen());

    return ReadFile(m_hMailslot, lpBuffer, nNumberOfBytesToRead, lpNumberOfBytesRead, nullptr);
  }

  bool GetInfo(_Out_opt_ LPDWORD lpMaxMessageSize, _Out_opt_ LPDWORD lpNextSize, _Out_opt_ LPDWORD lpMessageCount, _Out_opt_ LPDWORD lpReadTimeout) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(IsOpen());

    return GetMailslotInfo(m_hMailslot, lpMaxMessageSize, lpNextSize, lpMessageCount, lpReadTimeout);
  }

  bool SetReadTimeout(_In_ DWORD dwReadTimeout) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(IsOpen());

    return SetMailslotInfo(m_hMailslot, dwReadTimeout);
  }

  _NODISCARD bool IsOpen() const noexcept
  {
    return (m_hMailslot != INVALID_HANDLE_VALUE);
  }

  operator HANDLE() noexcept
  {
    return m_hMailslot;
  }

protected:
//Member variables
  HANDLE m_hMailslot;
};


class CMAILSLOT_EXT_CLASS CClientMailslot
{
public:
//Constructors / Destructors
  CClientMailslot() noexcept : m_hMailslot{INVALID_HANDLE_VALUE}
  {
  }
  CClientMailslot(const CClientMailslot&) = delete;
  CClientMailslot(CClientMailslot&&) = delete;
  ~CClientMailslot() noexcept
  {
    Close();
  }

//Methods
  CClientMailslot& operator=(const CClientMailslot&) = delete;
  CClientMailslot& operator=(CClientMailslot&&) = delete;

  bool Open(_In_ LPCTSTR lpszName, _In_opt_ LPSECURITY_ATTRIBUTES lpSecurityAttributes = nullptr) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(!IsOpen());

    m_hMailslot = CreateFile(lpszName, GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, lpSecurityAttributes, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, nullptr);
    return (m_hMailslot != INVALID_HANDLE_VALUE);
  }

  bool Close() noexcept
  {
    //What will be the return value from this method (assume the best)
    bool bSuccess{true};

    if (IsOpen())
    {
      bSuccess = CloseHandle(m_hMailslot);
      m_hMailslot = INVALID_HANDLE_VALUE;
    }

    return bSuccess;
  }

  bool Write(_In_reads_bytes_opt_(nNumberOfBytesToWrite) LPCVOID lpBuffer, _In_ DWORD nNumberOfBytesToWrite, _Out_opt_ LPDWORD lpNumberOfBytesWritten) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(IsOpen());

    return WriteFile(m_hMailslot, lpBuffer, nNumberOfBytesToWrite, lpNumberOfBytesWritten, nullptr);
  }

  _NODISCARD bool IsOpen() const noexcept
  {
    return (m_hMailslot != INVALID_HANDLE_VALUE);
  }

  operator HANDLE() noexcept
  {
    return m_hMailslot;
  }

protected:
//Member variables
  HANDLE m_hMailslot;
};

#endif //#ifndef __MAILSLOT_H__
